import weakref
from base64 import urlsafe_b64decode, urlsafe_b64encode
from pathlib import Path
from queue import SimpleQueue
from threading import Thread
from typing import Iterator, Optional, Tuple

from opentelemetry.proto.trace.v1.trace_pb2 import TracesData
from typing_extensions import TypeAlias

from phoenix.utilities.project import get_project_name

_Queue: TypeAlias = "SimpleQueue[Optional[TracesData]]"

_END_OF_QUEUE = None
_DIR_PREFIX = "project."


class TextFileSpanStoreImpl:
    def __init__(self, root: Path) -> None:
        self._root = root
        self._projects = dict(_load_projects(self._root))

    def save(self, traces_data: TracesData) -> None:
        for resource_spans in traces_data.resource_spans:
            name = get_project_name(resource_spans.resource.attributes)
            if (project := self._projects.get(name)) is None:
                self._projects[name] = project = _Project(name, self._root)
            project.save(TracesData(resource_spans=[resource_spans]))

    def load(self) -> Iterator[TracesData]:
        queue: _Queue = SimpleQueue()
        Thread(target=self._load_traces_data, args=(queue,)).start()
        while (item := queue.get()) is not _END_OF_QUEUE:
            yield item

    def _load_traces_data(self, queue: _Queue) -> None:
        """Load traces data from all projects into the queue"""
        for project in self._projects.values():
            project.load(queue)
        queue.put(_END_OF_QUEUE)


class _Project:
    def __init__(self, name: str, root: Path) -> None:
        self._path = root / f"project.{_b64encode(name.encode())}"
        spans_path = self._path / "spans"
        spans_path.mkdir(parents=True, exist_ok=True)
        self._spans = _Spans(spans_path / "spans.txt")

    def save(self, traces_data: TracesData) -> None:
        self._spans.save(traces_data)

    def load(self, queue: _Queue) -> None:
        self._spans.load(queue)


class _Spans:
    def __init__(self, file_path: Path):
        self._path = file_path
        self._file = self._path.open("a")
        weakref.finalize(self, self._file.close)

    def save(self, traces_data: TracesData) -> None:
        self._file.write(_b64encode(traces_data.SerializeToString()))
        self._file.write("\n")

    def load(self, queue: _Queue) -> None:
        with self._path.open("r") as f:
            while line := f.readline():
                queue.put(TracesData.FromString(_b64decode(line)))


def _load_projects(root: Path) -> Iterator[Tuple[str, _Project]]:
    for dir_path in root.glob(f"{_DIR_PREFIX}*/"):
        name = _b64decode(dir_path.name[len(_DIR_PREFIX) :]).decode()
        yield name, _Project(name, root)


def _b64encode(s: bytes) -> str:
    return urlsafe_b64encode(s).decode()


def _b64decode(name: str) -> bytes:
    return urlsafe_b64decode(name.encode())



from pathlib import Path
from typing import Callable, Iterator, Mapping, Protocol

from opentelemetry.proto.trace.v1.trace_pb2 import TracesData

from phoenix.config import SpanStorageType
from phoenix.storage.span_store.text_file import TextFileSpanStoreImpl


class SpanStore(Protocol):
    def save(self, req: TracesData) -> None: ...

    def load(self) -> Iterator[TracesData]: ...


SPAN_STORE_FACTORIES: Mapping[SpanStorageType, Callable[[Path], SpanStore]] = {
    SpanStorageType.TEXT_FILES: TextFileSpanStoreImpl,
}

__all__ = (
    "SpanStore",
    "SPAN_STORE_FACTORIES",
)



import logging
import weakref
from datetime import datetime
from io import BytesIO
from typing import List, Optional, Union, cast
from urllib.parse import urljoin

import pandas as pd
import pyarrow as pa
from pyarrow import ArrowInvalid
from requests import Session

import phoenix as px
from phoenix.config import (
    get_env_collector_endpoint,
    get_env_host,
    get_env_port,
    get_env_project_name,
)
from phoenix.session.data_extractor import TraceDataExtractor
from phoenix.trace import Evaluations
from phoenix.trace.dsl import SpanQuery

logger = logging.getLogger(__name__)


class Client(TraceDataExtractor):
    def __init__(
        self,
        *,
        endpoint: Optional[str] = None,
        use_active_session_if_available: bool = True,
    ):
        """
        Client for connecting to a Phoenix server.

        Args:
            endpoint (str, optional): Phoenix server endpoint, e.g. http://localhost:6006. If not
                provided, the endpoint will be inferred from the environment variables.
            use_active_session_if_available (bool, optional): If px.active_session() is available
                in the same runtime, e.g. the same Jupyter notebook, delegate the request to the
                active session instead of making HTTP requests. This argument is set to False if
                endpoint is provided explicitly.
        """
        self._use_active_session_if_available = use_active_session_if_available and not endpoint
        host = get_env_host()
        if host == "0.0.0.0":
            host = "127.0.0.1"
        self._base_url = (
            endpoint or get_env_collector_endpoint() or f"http://{host}:{get_env_port()}"
        )
        self._session = Session()
        weakref.finalize(self, self._session.close)
        if not (self._use_active_session_if_available and px.active_session()):
            self._warn_if_phoenix_is_not_running()

    def query_spans(
        self,
        *queries: SpanQuery,
        start_time: Optional[datetime] = None,
        stop_time: Optional[datetime] = None,
        root_spans_only: Optional[bool] = None,
        project_name: Optional[str] = None,
    ) -> Optional[Union[pd.DataFrame, List[pd.DataFrame]]]:
        """
        Queries spans from the Phoenix server or active session based on specified criteria.

        Args:
            queries (SpanQuery): One or more SpanQuery objects defining the query criteria.
            start_time (datetime, optional): The start time for the query range. Default None.
            stop_time (datetime, optional): The stop time for the query range. Default None.
            root_spans_only (bool, optional): If True, only root spans are returned. Default None.
            project_name (str, optional): The project name to query spans for. This can be set
                using environment variables. If not provided, falls back to the default project.

        Returns:
            Union[pd.DataFrame, List[pd.DataFrame]]: A pandas DataFrame or a list of pandas
                DataFrames containing the queried span data, or None if no spans are found.
        """
        project_name = project_name or get_env_project_name()
        if not queries:
            queries = (SpanQuery(),)
        if self._use_active_session_if_available and (session := px.active_session()):
            return session.query_spans(
                *queries,
                start_time=start_time,
                stop_time=stop_time,
                root_spans_only=root_spans_only,
                project_name=project_name,
            )
        response = self._session.get(
            url=urljoin(self._base_url, "/v1/spans"),
            json={
                "queries": [q.to_dict() for q in queries],
                "start_time": _to_iso_format(start_time),
                "stop_time": _to_iso_format(stop_time),
                "root_spans_only": root_spans_only,
                "project_name": project_name,
            },
        )
        if response.status_code == 404:
            logger.info("No spans found.")
            return None
        elif response.status_code == 422:
            raise ValueError(response.content.decode())
        response.raise_for_status()
        source = BytesIO(response.content)
        results = []
        while True:
            try:
                with pa.ipc.open_stream(source) as reader:
                    results.append(reader.read_pandas())
            except ArrowInvalid:
                break
        if len(results) == 1:
            df = results[0]
            return None if df.shape == (0, 0) else df
        return results

    def get_evaluations(
        self,
        project_name: Optional[str] = None,
    ) -> List[Evaluations]:
        """
        Retrieves evaluations for a given project from the Phoenix server or active session.

        Args:
            project_name (str, optional): The name of the project to retrieve evaluations for.
                This can be set using environment variables. If not provided, falls back to the
                default project.

        Returns:
            List[Evaluations]: A list of Evaluations objects containing evaluation data. Returns an
                empty list if no evaluations are found.
        """
        project_name = project_name or get_env_project_name()
        if self._use_active_session_if_available and (session := px.active_session()):
            return session.get_evaluations(project_name=project_name)
        response = self._session.get(
            urljoin(self._base_url, "/v1/evaluations"),
            json={"project_name": project_name},
        )
        if response.status_code == 404:
            logger.info("No evaluations found.")
            return []
        elif response.status_code == 422:
            raise ValueError(response.content.decode())
        response.raise_for_status()
        source = BytesIO(response.content)
        results = []
        while True:
            try:
                with pa.ipc.open_stream(source) as reader:
                    results.append(Evaluations.from_pyarrow_reader(reader))
            except ArrowInvalid:
                break
        return results

    def _warn_if_phoenix_is_not_running(self) -> None:
        try:
            self._session.get(urljoin(self._base_url, "/arize_phoenix_version")).raise_for_status()
        except Exception:
            logger.warning(
                f"Arize Phoenix is not running on {self._base_url}. Launch Phoenix "
                f"with `import phoenix as px; px.launch_app()`"
            )

    def log_evaluations(self, *evals: Evaluations, project_name: Optional[str] = None) -> None:
        """
        Logs evaluation data to the Phoenix server.

        Args:
            evals (Evaluations): One or more Evaluations objects containing the data to log.
            project_name (str, optional): The project name under which to log the evaluations.
                This can be set using environment variables. If not provided, falls back to the
                default project.

        Returns:
            None
        """
        project_name = project_name or get_env_project_name()
        for evaluation in evals:
            table = evaluation.to_pyarrow_table()
            sink = pa.BufferOutputStream()
            headers = {"content-type": "application/x-pandas-arrow"}
            if project_name:
                headers["project-name"] = project_name
            with pa.ipc.new_stream(sink, table.schema) as writer:
                writer.write_table(table)
            self._session.post(
                urljoin(self._base_url, "/v1/evaluations"),
                data=cast(bytes, sink.getvalue().to_pybytes()),
                headers=headers,
            ).raise_for_status()


def _to_iso_format(value: Optional[datetime]) -> Optional[str]:
    return value.isoformat() if value else None



import weakref
from collections import defaultdict
from queue import SimpleQueue
from threading import RLock, Thread
from types import MethodType
from typing import DefaultDict, Iterator, Optional, Tuple, Union

from typing_extensions import assert_never

import phoenix.trace.v1 as pb
from phoenix.config import DEFAULT_PROJECT_NAME
from phoenix.core.project import (
    END_OF_QUEUE,
    Project,
    _ProjectName,
)
from phoenix.trace.schemas import Span

_SpanItem = Tuple[Span, _ProjectName]
_EvalItem = Tuple[pb.Evaluation, _ProjectName]


class Traces:
    def __init__(self) -> None:
        self._span_queue: "SimpleQueue[Optional[_SpanItem]]" = SimpleQueue()
        self._eval_queue: "SimpleQueue[Optional[_EvalItem]]" = SimpleQueue()
        # Putting `None` as the sentinel value for queue termination.
        weakref.finalize(self, self._span_queue.put, END_OF_QUEUE)
        weakref.finalize(self, self._eval_queue.put, END_OF_QUEUE)
        self._lock = RLock()
        self._projects: DefaultDict[_ProjectName, "Project"] = defaultdict(
            Project,
            {DEFAULT_PROJECT_NAME: Project()},
        )
        self._start_consumers()

    def get_project(self, project_name: str) -> Optional["Project"]:
        with self._lock:
            return self._projects.get(project_name)

    def get_projects(self) -> Iterator[Tuple[int, str, "Project"]]:
        with self._lock:
            for project_id, (project_name, project) in enumerate(self._projects.items()):
                if project.is_archived:
                    continue
                yield project_id, project_name, project

    def archive_project(self, id: int) -> Optional["Project"]:
        with self._lock:
            active_projects = {
                project_id: project
                for project_id, _, project in self.get_projects()
                if not project.is_archived
            }
            if len(active_projects) <= 1:
                return None
            if project := active_projects.get(id):
                project.archive()
                return project
        return None

    def put(
        self,
        item: Union[Span, pb.Evaluation],
        project_name: Optional[str] = None,
    ) -> None:
        if not project_name:
            project_name = DEFAULT_PROJECT_NAME
        if isinstance(item, Span):
            self._span_queue.put((item, project_name))
        elif isinstance(item, pb.Evaluation):
            self._eval_queue.put((item, project_name))
        else:
            assert_never(item)

    def _start_consumers(self) -> None:
        Thread(
            target=MethodType(self.__class__._consume_spans, weakref.proxy(self)),
            args=(self._span_queue,),
            daemon=True,
        ).start()
        Thread(
            target=MethodType(self.__class__._consume_evals, weakref.proxy(self)),
            args=(self._eval_queue,),
            daemon=True,
        ).start()

    def _consume_spans(self, queue: "SimpleQueue[Optional[_SpanItem]]") -> None:
        while (item := queue.get()) is not END_OF_QUEUE:
            span, project_name = item
            with self._lock:
                project = self._projects[project_name]
            project.add_span(span)

    def _consume_evals(self, queue: "SimpleQueue[Optional[_EvalItem]]") -> None:
        while (item := queue.get()) is not END_OF_QUEUE:
            pb_eval, project_name = item
            with self._lock:
                project = self._projects[project_name]
            project.add_eval(pb_eval)



from typing import Optional

from phoenix.config import get_env_span_storage_type, get_storage_dir
from phoenix.core.traces import Traces
from phoenix.storage.span_store import SPAN_STORE_FACTORIES, SpanStore
from phoenix.trace.otel import decode
from phoenix.utilities.project import get_project_name


def get_span_store() -> Optional[SpanStore]:
    if span_store_type := get_env_span_storage_type():
        span_store_factory = SPAN_STORE_FACTORIES[span_store_type]
        return span_store_factory(get_storage_dir())
    return None


def load_traces_data_from_store(traces: Traces, span_store: SpanStore) -> None:
    for traces_data in span_store.load():
        for resource_spans in traces_data.resource_spans:
            project_name = get_project_name(resource_spans.resource.attributes)
            for scope_span in resource_spans.scope_spans:
                for span in scope_span.spans:
                    traces.put(decode(span), project_name=project_name)
